"""
:copyright: (c) 2017 by Jumper Labs Ltd.
:license: Apache 2.0, see LICENSE.txt for more details.
"""
from __future__ import print_function
from builtins import input
import os
import errno
import subprocess
import hashlib
import sys
from time import sleep
from shutil import copyfile
from shutil import copymode
import requests
from distutils.version import LooseVersion
from __version__ import __version__ as jumper_current_version
import json
from termcolor import colored
from terminaltables import SingleTable
from threading import Thread

import timeout_decorator

from .jemu_uart import JemuUart
from .jemu_peripherals_parser import JemuPeripheralsParser
from .jemu_bsp_parser import JemuBspParser
from .jemu_gpio import JemuGpio
from .jemu_connection import JemuConnection
from .jemu_web_api import JemuWebApi
from .jemu_interrupts import JemuInterrupts
from .common import TimeoutException
import tarfile

config_file_name = 'config.json' if 'JUMPER_STAGING' not in os.environ else 'config.staging.json'
DEFAULT_CONFIG = os.path.join(os.path.expanduser('~'), '.jumper', config_file_name)


class VlabException(Exception):
    def __init__(self, message, exit_code):
        super(VlabException, self).__init__(message)
        self.exit_code = exit_code
        self.message = message


class EmulationError(VlabException):
    def __init__(self, message):
        super(EmulationError, self).__init__(message, 5)


class MissingFileError(VlabException):
    def __init__(self, message):
        super(MissingFileError, self).__init__(message, 2)


class ArgumentError(VlabException):
    def __init__(self, message):
        super(ArgumentError, self).__init__(message, 1)


class TranspilerError(VlabException):
    def __init__(self, message):
        super(TranspilerError, self).__init__(message, 3)

class Vlab(object):
    """
    The main class for using Jumper Virtual Lab

    :param working_directory: The directory that holds the bsp.json abd scenario.json files for the virtual session
    :param config_file: Config file holding the API token (downloaded from https://vlab.jumper.io)
    :param gdb_mode: If True, a GDB server will be opened on port 5555
    :param sudo_mode: If True, firmware can write to read-only registers. This is useful for injecting a mock state to the hardware.
    :param print_trace: If ture, a trace with the registers value at each tick of the execution will be generated.
    :param trace_output_file: If print_trace is True, sets the trace file. Default is stdout.
    :param print_uart: If True UART prints coming from the device will be printed to stdout or a file
    :param uart_output_file: If print_uart is True, sets the UART output file. Default is stdout.
    """
    if os.environ.get('JEMU_DIR') is None:
        _transpiler_dir = os.path.abspath(os.path.join(os.path.dirname(__file__), '..', '..', 'jemu'))
    else:
        _transpiler_dir = os.path.abspath(os.environ['JEMU_DIR'])

    _jemu_build_dir = os.path.abspath(os.path.join(_transpiler_dir, 'emulator', '_build'))
    _jemu_bin_src = os.path.join(_jemu_build_dir, 'jemu')
    _jemu_bin_tgz_src = os.path.join(_jemu_build_dir, 'jemu.tgz')
    _local_cache_file = os.path.abspath(os.path.join(_transpiler_dir, 'emulator', '_build','jemu.cache.sha1'))

    _INT_TYPE = "interrupt_type"

    _TYPE_STRING = "type"
    _BKPT = "bkpt"
    _VALUE_STRING = "value"

    _examples_hash_list = [
        b'f9cf392be5ac94b89e0b0837f258a6c177efb2be',
        b'1ef133cfe2a956cf6a61f629f818c2dbfd866367',
        b'be24b36eab847ea1268c21a75f992d5a63f91d4b',
        b'bcf18382d0c130e5f34e8053bb3ab56493d6c589',
        b'16572d43ccfdcef0e769fc934daed287946c1156']

    @staticmethod
    def _get_latest_version(name):
        url = "https://pypi.python.org/pypi/{}/json".format(name)
        try:
            return list(reversed(sorted(requests.get(url).json()["releases"], key=LooseVersion)))[0]
        except Exception as e:
            return None

    @staticmethod
    def _print_upate_to_screen(jumper_latest_version, jumper_current_version):
        update_message = "Update available {0} ".format(jumper_current_version) + u'\u2192' + colored(" " + jumper_latest_version, 'green', attrs=['bold'])
        how_to_updtae_message = "\n  Run " + colored(" sudo pip install jumper --upgrade ", "blue", attrs=['bold']) + "to update"
        table_data = [[update_message + how_to_updtae_message]]
        table = SingleTable(table_data)
        table.padding_left = 2
        table.padding_right = 2
        print()
        print(table.table.encode('utf-8'))
        print()

    @staticmethod
    def check_version():
        jumper_latest_version = Vlab._get_latest_version("jumper")
        if jumper_latest_version:
            if LooseVersion(jumper_current_version) < LooseVersion(jumper_latest_version):
                Vlab._print_upate_to_screen(jumper_latest_version, jumper_current_version)

    def is_jemu_latest_version(self):
        if self._local_jemu:
            return True
        jumper_latest_version = self._web_api.get_jemu_version()
        if jumper_latest_version and os.path.isfile(self._jemu_bin):
            jemu_cmd = [self._jemu_bin, '-v']
            try:
                jemu_version = subprocess.check_output(jemu_cmd, cwd=self._working_directory)
                jemu_version = jemu_version.rstrip()
            except Exception as e:
                return False
            result = LooseVersion(jemu_version) == LooseVersion(jumper_latest_version)
            return result
        return False

    def __init__(self, working_directory=None, config_file=None, gdb_mode=False, sudo_mode=False, print_trace=False, trace_output_file="", print_uart=False, uart_output_file=""):
        self.check_version()
        self._working_directory = os.path.abspath(working_directory) if working_directory else self._transpiler_dir
        self._local_jemu = True if (('LOCAL_JEMU' in os.environ) or ('JEMU_LOCAL' in os.environ)) else False
        self._gdb_mode = gdb_mode
        self._sudo_mode = sudo_mode
        self._jemu_process = None
        self._was_start = False
        self._uart_device_path = os.path.join(self._working_directory, 'uart')
        self._jemu_server_address = "localhost"
        self._jemu_bin = os.path.join(self._working_directory, 'jemu')
        self._jemu_bin_tgz = os.path.join(self._working_directory, 'jemu.tgz')
        self._cache_file = self._jemu_bin + ".cache.sha1"
        self._uart = JemuUart(self._uart_device_path, self)
        self._uart.remove()
        self._jemu_connection = JemuConnection(self._jemu_server_address)
        self._print_trace = print_trace
        self._trace_output_file = trace_output_file
        self._print_uart = print_uart
        self._uart_output_file = uart_output_file
        self._jemu_debug = True if 'JEMU_DEBUG' in os.environ else False
        self._on_bkpt = None
        self._return_code = None
        self._jemu_gpio = JemuGpio()
        self._jemu_interrupt = None
        self._stdout_thread = None
        self._build_components_methods()
        self.jemu_server_port = 8000
        token = None

        config_file = config_file or DEFAULT_CONFIG

        if not os.path.isfile(config_file):
            raise MissingFileError('Config file not found at: {}'.format(os.path.abspath(config_file)))

        with open(config_file) as config_data:
            config = json.load(config_data)
        if 'token' in config:
            token = config['token']

        if not self._local_jemu:
            if token is not None:
                self._web_api = JemuWebApi(jumper_token=token)
            else:
                raise MissingFileError('Config file not found at: {}'.format(config_file))

    @staticmethod
    def _silent_remove_file(filename):
        try:
            if os.path.isfile(filename):
                os.remove(filename)
        except OSError as e:
            if e.errno != errno.ENOENT:
                raise

    def _valid_file_existence(self, file_path):
        if not os.path.isfile(file_path):
            raise MissingFileError("Failed to open binary file (at: '" + file_path + "')")

    @property
    def uart(self):
        """
        The main UART device for the Virtual Lab session

        :return: :class:`~jumper.jemu_uart.JemuUart`
        """
        return self._uart

    @property
    def gpio(self):
        return self._jemu_gpio

    @property
    def interrupts(self):
        return self._jemu_interrupt

    # @property
    # def interrupt_type(self):
    #     return self._INT_TYPE

    def _build_components_methods(self):
        peripherals_json = os.path.join(self._working_directory, "peripherals.json")
        bsp_json = os.path.join(self._working_directory, "bsp.json")
        default_bsp_json = os.path.join(os.path.dirname(__file__), "default_bsp.json")

        if os.path.isfile(bsp_json):
            components_list = self._parse_bsp_json(bsp_json)
        elif os.path.isfile(peripherals_json):
            components_list = self._parse_peripherals_json(peripherals_json)
        else:
            components_list = self._parse_bsp_json(default_bsp_json)

        for component in components_list:
            setattr(self, component["name"], component["obj"])

    def _parse_bsp_json(self, bsp_path):
        bsp_json_parser = JemuBspParser(bsp_path)
        return bsp_json_parser.get_components(self._jemu_connection)

    def _parse_peripherals_json(self, peripherals_path):
        """Backwards compatibility"""
        self._peripherals_json_parser = \
            JemuPeripheralsParser(os.path.join(self._working_directory, peripherals_path))
        return self._peripherals_json_parser.get_peripherals(self._jemu_connection)

    @staticmethod
    def _get_file_signature(file_path):
        sha1 = hashlib.sha1()

        with open(file_path, 'rb') as f:
            while True:
                data = f.read(65536)
                if not data:
                    break
                sha1.update(data)

        return sha1.hexdigest()

    def _read_file_signature_backup(self, file_path):
        data = ''
        if os.path.isfile(file_path):
            if os.path.isfile(self._jemu_bin):
                with open(file_path, 'r') as f:
                    data = f.read().replace('\n', '')
            else:
                os.remove(file_path)

        return data

    def _write_file_signature_backup(self, sha1_cache_string, file_path):
        with open(file_path, 'w+') as f:
            f.write(sha1_cache_string)

    def _unzip_jemu(self, gen_new):
        if gen_new:
            try:
                tar = tarfile.open(self._jemu_bin_tgz)
                tar.extractall(path=self._working_directory)
                tar.close()
            except Exception:
                raise VlabException('Error: Could not open jemu file')

            os.chmod(self._jemu_bin, 0o777)
            os.remove(self._jemu_bin_tgz)

    def load(self, file_path):
        """
        Loads firmware to a virtual device and initialises a Virtual Lab session.
        Use :func:`~jumper.Vlab.start()` to start an emulation after this method was called.

        :param file_path: Path for a firmware file (supported extensions are bin, out, elf, hex)
        """

        ext = [".bin", ".out", ".elf", ".hex"]
        if not file_path.endswith(tuple(ext)):
            raise ArgumentError('Invalid file extension - supported extensions are bin, out, elf, hex')

        self._valid_file_existence(file_path)
        file_path = os.path.abspath(file_path)
        filename = os.path.basename(file_path)
        gen_new = True
        new_signature = self._get_file_signature(file_path)

        prev_signature = self._read_file_signature_backup(self._cache_file)
        if prev_signature == new_signature and self.is_jemu_latest_version():
            if not self._local_jemu:
                gen_new = False
            else:
                local_signature = self._read_file_signature_backup(self._local_cache_file)
                if local_signature == new_signature:
                    gen_new = False

        if gen_new:
            self._silent_remove_file(self._jemu_bin)
            self._silent_remove_file(self._cache_file)
            if self._local_jemu:
                self._silent_remove_file(self._local_cache_file)

        if self._local_jemu:
            try:
                if gen_new:
                    transpiler_cmd = ["node", "index.js", "--debug", "--platform", "nrf52", "--bin", file_path, "--zip"]
                    subprocess.check_call(transpiler_cmd, cwd=self._transpiler_dir, stdout=open(os.devnull, 'w'), stderr=None)
                    copyfile(self._jemu_bin_tgz_src, self._jemu_bin_tgz)
                else:  # from cache
                    make_cmd = ['make', '-C', 'emulator', 'DEBUG=1', 'PLATFORM=nrf52', '-j8']
                    subprocess.check_call(make_cmd, cwd=self._transpiler_dir, stdout=open(os.devnull, 'w'), stderr=None)
                    copyfile(self._jemu_bin_src, self._jemu_bin)
                    copymode(self._jemu_bin_src, self._jemu_bin)

            except subprocess.CalledProcessError as e:
                raise TranspilerError("Transpiler failed with an error: " + e.message)

        else:  # from cloud
            if gen_new:
                if new_signature in self._examples_hash_list:
                    self._web_api.send_event('upload example firmware')
                else:
                    self._web_api.send_event('upload new firmware')

                with open(file_path, 'r') as data:
                    self._web_api.get_archived_emulator(filename, data, self._jemu_bin_tgz)
            else:  # from cache
                self._web_api.send_event('using cached firmware')

        self._unzip_jemu(gen_new)

        if gen_new and os.path.isfile(self._jemu_bin):
            self._write_file_signature_backup(new_signature, self._cache_file)
            if self._local_jemu:
                copyfile(self._cache_file, self._local_cache_file)
                copymode(self._cache_file, self._local_cache_file)

    def start(self, ns=None):
        """
        Starts the emulation

        :param ns: If provided, commands the virtual device to run for the amount of time given in ns and then halt.

            If this parameter is used, this function is blocking until the virtual devices halts,
            if None is given, this function is non-blocking.
        """
        if not os.path.isfile(self._jemu_bin):
            raise MissingFileError(self._jemu_bin + ' is not found')
        elif not os.access(self._jemu_bin, os.X_OK):
            raise MissingFileError(self._jemu_bin + ' is not executable')

        self._was_start = True

        jemu_cmd = []
        # if self._jemu_debug:
        #     jemu_cmd = ['gdbserver', 'localhost:7777']
        jemu_cmd.append(self._jemu_bin)
        jemu_cmd.append('-w')
        jemu_cmd.append('--sdk-port')
        jemu_cmd.append('0')

        if self._gdb_mode:
            jemu_cmd.append('-g')
        if self._sudo_mode:
            jemu_cmd.append('-s')
        if self._print_trace:
            jemu_cmd.append('-t')
            if self._trace_output_file != "":
                jemu_cmd.append(self._trace_output_file)
        if self._print_uart:
            jemu_cmd.append('-u')
            if self._uart_output_file != "":
                jemu_cmd.append(self._uart_output_file)

        def jemu_connection():
            @timeout_decorator.timeout(6)
            def wait_for_connection():
                while not self._jemu_connection.connect(self.jemu_server_port):
                    sleep(0.1)
            try:
                wait_for_connection()
            except timeout_decorator.TimeoutError:
                self.stop()
                raise EmulationError("Error: Couldn't connect to Emulator")
            if not self._jemu_connection.handshake(ns):
                raise EmulationError("Error: Couldn't connect to Emulator")

            self._jemu_gpio.set_connection_manager(self._jemu_connection)
            self._jemu_interrupt = JemuInterrupts(self._jemu_connection)
            self._jemu_connection.register(self.receive_packet)

        if self._jemu_debug:
            input("Start a debugger with the following parameters:\n\
            cwd: {}\n\
            command: {}\n\
            Press Enter to continue...".format(self._working_directory, ' '.join(jemu_cmd))
                  )
            jemu_connection()
        else:
            try:
                self._jemu_process = subprocess.Popen(jemu_cmd, cwd=self._working_directory, stdout=subprocess.PIPE)
                sleep(0.3)
            except Exception as e:
                raise EmulationError(e.message)

            self.jemu_server_port = self.get_port_number()
            jemu_connection()
            self._stdout_thread = Thread(target=self._thread_stdout_function)
            self._stdout_thread.start()

        @timeout_decorator.timeout(6)
        def wait_for_uart():
            while not os.path.exists(self._uart_device_path):
                sleep(0.1)

        try:
            wait_for_uart()
        except timeout_decorator.TimeoutError:
            if not self.is_running():
                raise EmulationError("Error: Emulator failed on start")
            else:
                self.stop()
                raise EmulationError("Error: Uart doesn't exist")

        self._uart.open()

        self._jemu_connection.start()

    def _thread_stdout_function(self):
        while True:
            data = self._jemu_process.stdout.read(1)
            if data == '':
                break
            sys.stdout.write(data)

    def get_port_number(self):
        @timeout_decorator.timeout(6)
        def get_next_stdout_line():
            return self._jemu_process.stdout.readline()

        loop_number = 0
        while loop_number < 10:
            try:
                line = get_next_stdout_line()
            except timeout_decorator.TimeoutError:
                raise EmulationError("Error: Could not read port from emulator")

            line_data = line.split(' ')
            length = len(line_data)
            for index in range(0, length):
                if line_data[index] == 'port:':
                    return line_data[index + 1].split('\n')[0]
            loop_number += 1

        raise EmulationError("Error: Could not read port from emulator")

    def stop(self):
        """
        Stops the Virtual Lab session.

        Opposing to halting the session, the virtual device cannot be resumed after a stop command.

        """
        if self._jemu_connection:
            self._jemu_connection.close()
        if self._uart:
            self._uart.close()
            self._uart.remove()

        if self._jemu_process and self._jemu_process.poll() is None:
            self._jemu_process.terminate()
            self._jemu_process.wait()
            self._return_code = 0

        self._stop_stdout_thread()
        self._uart = None
        self._jemu_connection = None

    def _stop_stdout_thread(self):
        if self._stdout_thread.is_alive():
            self._stdout_thread.join()

    def run_for_ms(self, ms):
        """
        Starts or resumes the virtual device, the device will halt after the amount of time specified.

        This function is blocking until the virtual device has halted. Use this when the virtual device is stopped
        or halted.

        :param ms: Time to run in ms
        """
        self.run_for_us(ms * 1000)

    def run_for_us(self, us):
        """
        Starts or resumes the virtual device, the device will halt after the amount of time specified.

        This function is blocking until the virtual device has halted. Use this when the virtual device is stopped
        or halted.

        :param ms: Time to run in us
        """
        self.run_for_ns(us * 1000)

    def run_for_ns(self, ns):
        """
        Starts or resumes the virtual device, the device will halt after the amount of time specified.

        This function is blocking until the virtual device has halted. Use this when the virtual device is stopped
        or halted.

        :param ns: Time to run in ns
        """
        if not self._was_start:
            self.start(ns)
            self.SUDO.wait_until_stopped()
        else:
            self.SUDO.run_for_ns(ns)

    def stop_after_ms(self, ms):
        # """
        # Causes the virtual device to halt after the amount of time specified.
        # This function is non-blocking and does not cause the device to resume.
        #
        # Use this when the virtual device is halted.
        #
        # :param ms: Time to run in ms
        # """
        self.stop_after_ns(ms * 1000000)

    def stop_after_ns(self, ns):
        # """
        # Causes the virtual device to halt after the amount of time specified.
        # This function is non-blocking and does not cause the device to resume.
        #
        # Use this when the virtual device is halted.
        #
        # :param ns: Time to run in ns
        # """
        self.SUDO.stop_after_ns(ns)

    def resume(self):
        """
        Resumes a paused device.

        """
        self.SUDO.resume()

    def cancel_stop(self):
        self.SUDO.cancel_stop()

    def pause(self):
        """
        Pause the device.

        """
        self.run_for_ns(0)

    def on_interrupt(self, callback):
        self.interrupts.on_interrupt([callback])

    def set_timer(self, ns, callback):
        self.SUDO.set_timer(ns, callback)

    def get_state(self):
        if not self._was_start:
            return "init"
        elif not self.is_running():
            return "stopped"
        return self.SUDO.get_state()

    def on_pin_level_event(self, callback):
        """
        Specifies a callback for a pin transition event.

        :param callback: The callback to be called when a pin transition occures. The callback will be called with callback(pin_number, pin_level)
        """
        self.gpio.on_pin_level_event(callback)

    def get_pin_level(self, pin_num):
        """
        Specifies get the pin level for a pin num.

        :param pin num: pin number id
        """
        return self.gpio.get_pin_level(pin_num)

    def on_bkpt(self, callback):
        """
        Sets a callback to be called when the virtual device execution reaches a BKPT assembly instruction.

        :param callback: The callback to be called. Callback will be called with callback(code)\
        where code is the code for the BKPT instruction.
        """
        self._on_bkpt = callback

    def receive_packet(self, jemu_packet):
        if jemu_packet[self._TYPE_STRING] == self._BKPT:
            if self._on_bkpt is not None:
                bkpt_code = jemu_packet[self._VALUE_STRING]
                self._on_bkpt(bkpt_code)

    def is_running(self):
        """
        Checks if the virtual device has been started.

        :return: True if running or pause, False otherwise.
        """
        if not self._jemu_process:
            if self._jemu_debug:
                return True
            else:
                return False

        self._return_code = self._jemu_process.poll()
        return self._return_code is None

    def get_return_code(self):
        """
        Checks a return code from the device.

        :return:
            - 0 if device was stopped using the :func:`~stop()` method
            - Exit code from firmware if the Device exited using the jumper_sudo_exit_with_exit_code() \
            command from jumper.h
        """
        if not self._jemu_process:
            return None

        if self._return_code is None:
            self._return_code = self._jemu_process.poll()

        return self._return_code

    def get_device_time_ns(self):
        """
        How much time passed from beginning of the emulation.

        :return: Emulation time in nanoseconds.
        """
        return self.SUDO.get_device_time_ns()

    def get_device_time_us(self):
        """
        How much time passed from beginning of the emulation.

        :return: Emulation time in microseconds.
        """
        return self.get_device_time_ns() / 1000

    def get_device_time_ms(self):
        """
        How much time passed from beginning of the emulation.

        :return: Emulation time in milliseconds.
        """
        return self.get_device_time_us() / 1000

    def __enter__(self):
        self.start()
        return self

    def __exit__(self, *err):
        self.stop()

    def __del__(self):
        self.stop()




"""
:copyright: (c) 2017 by Jumper Labs Ltd.
:license: Apache 2.0, see LICENSE.txt for more details.
"""
from threading import Thread
import socket
import json
import struct
import logging
import sys

class JemuConnection:
    _addr = None
    _conn = None
    _HANDSHAKE = "handshake"
    _COMMAND = "command"
    _COMMAND_TYPE = "command_type"
    _PERIPHERAL_ID = "peripheral_id"
    _TRANSITION_TYPE = "transition_type"
    _PIN_NUM = "pin_number"
    _MESSAGE_ID = "message_id"
    _VOLTAGE = "voltage"
    _API_VERSION = "1"
    _OK = "ok"
    _ERROR = "error"
    _ERROR_CODE = "404"
    _TYPE = "type"
    _RESPONSE = "response"
    _API_V_STRING = "api_version"
    _ERROR_MESSAGE = "error"
    _COMMAND_START = "start"
    _NANO_SECONDS = "nanoseconds"

    _LOG_LEVEL = logging.ERROR
    _logger = logging.getLogger()
    log_handler = logging.StreamHandler(sys.stdout)
    formatter = logging.Formatter('%(asctime)s - %(levelname)s - %(message)s')
    log_handler.setFormatter(formatter)
    _logger.addHandler(log_handler)
    _logger.setLevel(_LOG_LEVEL)
    log_handler.setLevel(_LOG_LEVEL)

    def __init__(self, addr):
        self._addr = addr
        self._conn = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self._conn.settimeout(0.3)
        self._callbacks = []
        self._thread = None
        self._should_run = False
        self._connection_closed = True
        self._threads = []

    def start(self):
        self._should_run = True
        t = Thread(target=self.connection_task)
        self._threads.append(t)
        t.start()

    def close(self):
        self._should_run = False
        for t in self._threads:
            t.join()

    def connection_task(self):
        while self._should_run:
            json_buff_string = self.recv_json()
            if json_buff_string is None:
                self._should_run = False
                break
            self._logger.debug("received data: " + json_buff_string)
            json_pack = json.loads(json_buff_string)
            t = Thread(target=self.inform, args=(json_pack,))
            self._threads.append(t)
            t.start()
            if len(self._threads) > 4:
                self.clean_threat_list()

        if not self._connection_closed:
            self._conn.close()
            self._connection_closed = True

    def inform(self, jemu_json_packet):
        for callback in self._callbacks:
            callback(jemu_json_packet)

    def register(self, callback):
        self._callbacks.append(callback)

    def connect(self, port):
        result = False
        try:
            self._conn.connect((self._addr, int(port)))
            self._connection_closed = False
            result = True
        except:
            self._conn = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        finally:
            return result

    def handshake(self, ns=None):
        self._should_run = True
        self.send_handshake()
        response_ack = self.recv_json()
        if response_ack is None:
            return False
        self.check_response(response_ack)
        self.send_start(ns)
        return True

    def check_response(self,response_ack):
        self._logger.debug("handshake response data: " + response_ack)
        response = json.loads(response_ack)
        message_type = response[self._TYPE]
        if message_type != self._RESPONSE:
            raise JemuConnectionException("Error expect response message instead got [" + message_type + "]")
        status_code = response[self._RESPONSE]
        if status_code != self._OK:
            raise JemuConnectionException("Sdk's api version does not match to the jemu's version")

    def send_json(self, json_obj):
        data_to_send = json.dumps(json_obj)
        number_of_bytes = len(data_to_send.encode('utf-8'))
        self._logger.debug("Number of byte of send data: " + str(number_of_bytes))
        number_to_send = struct.pack('!i', number_of_bytes)
        self.send(number_to_send)
        self._logger.debug("Data to send: " + data_to_send)
        self.send(data_to_send)

    def send_handshake(self):
        hand_shake_message = {self._API_V_STRING: self._API_VERSION, self._TYPE: self._HANDSHAKE}
        self.send_json(hand_shake_message)

    def send_start(self, ns=None):
        ack = {self._TYPE: self._COMMAND, self._COMMAND: self._COMMAND_START}
        if ns:
            ack[self._NANO_SECONDS] = ns
        self.send_json(ack)

    def send(self, buffer):
        try:
            self._conn.sendall(buffer)
        except socket.error as e:
            raise JemuConnectionException(str(e))

    def clean_threat_list(self):
        for t in self._threads:
            if not t.isAlive():
                self._threads.remove(t)

    def recv_json(self):
        data = self.receive(4)
        if data is None:
            return None
        buffer_size = struct.unpack("!i", data)[0]
        self._logger.debug("Number of byte of received data: " + str(buffer_size))
        if buffer_size == 0:
            return None
        data = self.receive(buffer_size)
        if data is None:
            return None
        else:
            return data

    def receive(self, size):
        result = b''
        while (len(result) < size) and (not self._connection_closed):
            if not self._should_run:
                return None

            try:
                data = self._conn.recv(size - len(result))
            except socket.timeout:
                continue
            except socket.error as e:
                if self._should_run:
                    raise JemuConnectionException(str(e))
                else:
                    self._connection_closed = True
                    return None

            if data == b'':
                self._connection_closed = True
                return None
            else:
                result += data

        return result


class JemuConnectionException(Exception):
    pass



"""
:copyright: (c) 2017 by Jumper Labs Ltd.
:license: Apache 2.0, see LICENSE.txt for more details.
"""
from threading import Lock


class JemuGpio:
    _PIN_NUM = "pin_number"
    _TRANSITION_TYPE = "transition_type"

    def __init__(self):
        self._pin_level_callback = None
        self._jemu_socket_manager = None
        self._pin_level_statues = dict()
        self._lock = Lock()

    def on_pin_level_event(self, callback):
        with self._lock:
            self._pin_level_callback = callback

    def receive_packet(self, jemu_packet):
        if jemu_packet["type"] == "pin_level_event":
            with self._lock:
                self._pin_level_statues[jemu_packet[self._PIN_NUM]] = jemu_packet[self._TRANSITION_TYPE]
                if self._pin_level_callback:
                    self._pin_level_callback(jemu_packet[self._PIN_NUM], jemu_packet[self._TRANSITION_TYPE])

    def set_connection_manager(self, jemu_socket_manager):
        self._jemu_socket_manager = jemu_socket_manager
        self._jemu_socket_manager.register(self.receive_packet)

    def get_pin_level(self, pin_num):
        with self._lock:
            if pin_num in self._pin_level_statues:
                return self._pin_level_statues[pin_num] 
            else:
                return 1



"""
:copyright: (c) 2017 by Jumper Labs Ltd.
:license: Apache 2.0, see LICENSE.txt for more details.
"""
from __future__ import print_function
import argparse
import signal
from vlab import VlabException
from jemu_web_api import WebException
from os import getcwd
from time import sleep
from threading import Event
from __version__ import __version__ as jumper_current_version

import sys

from .vlab import Vlab
from .vlab_hci_device import VirtualHciDevice
from . import __version__


class _VersionAction(argparse.Action):
    def __init__(self,
                 option_strings,
                 version=None,
                 dest=argparse.SUPPRESS,
                 default=argparse.SUPPRESS,
                 help="show program's version number and exit"):
        super(_VersionAction, self).__init__(
            option_strings=option_strings,
            dest=dest,
            default=default,
            nargs=0,
            help=help)
        self.version = version

    def __call__(self, parser, namespace, values, option_string=None):
        version = self.version
        if version is None:
            version = parser.version
        formatter = parser._get_formatter()
        formatter.add_text(version)
        Vlab.check_version()
        parser.exit(message=formatter.format_help())


def run(args):
    try:
        vlab = Vlab(
            working_directory=args.directory,
            sudo_mode=args.sudo,
            gdb_mode=args.gdb,
            print_trace=args.trace,
            trace_output_file=args.trace_dest,
            print_uart=args.uart,
            uart_output_file=args.uart_dest
        )
        vlab.load(args.bin)

        reached_bkpt = Event()

        def bkpt_callback(code):
            print('Firmware reached a BKPT instruction with code {}'.format(code))
            reached_bkpt.set()

        vlab.on_bkpt(bkpt_callback)

        def pins_listener(pin_number, pin_level):
            print("pin_number:", pin_number, "  pin_level:", pin_level)

        if args.gpio:
            print("gpio")
            vlab.on_pin_level_event(pins_listener)


        vlab.start()

        if (not args.uart) and (not args.trace):
            print('\nVirtual device is running without UART/Trace prints (use -u and/or -t to get your firmware execution status)')
        else:
            print('\nVirtual device is running')

        while (not reached_bkpt.is_set()) and vlab.is_running():
            sleep(0.1)

        if reached_bkpt and vlab.is_running():
            vlab.stop()

        if vlab.get_return_code():
            vlab.stop()
            sys.exit(vlab.get_return_code())

    except VlabException as e:
        print(e.message)
        try:
            vlab.stop()
        finally:
            exit(e.exit_code)


def hci():
    virtual_hci_device = VirtualHciDevice()
    virtual_hci_device.start()
    try:
        while True:
            sleep(0.1)
    except KeyboardInterrupt:
        pass
    finally:
        virtual_hci_device.stop()


def signal_handler(signal, frame):
    sys.exit(1)


def main():
    signal.signal(signal.SIGINT, signal_handler)
    parser = argparse.ArgumentParser(
        prog='jumper',
        description="CLI interface for using Jumper's emulator"
    )

    parser.add_argument('--version', action=_VersionAction, version='%(prog)s {}'.format(__version__))

    subparsers = parser.add_subparsers(title='Commands', dest='command')

    run_parser = subparsers.add_parser(
        'run',
        help='Runs an emulator with a binary FW file. Currently only support nRF52 devices'
    )
    run_parser.add_argument(
        '--bin',
        '-b ',
        '--fw',
        help="Firmware to be flashed to the virtual device (supported extensions are bin, out, elf, hex). In case more than one file needs to be flashed (such as using Nordic's softdevice), the files should be merged first. Check out https://vlab.jumper.io/docs#softdevice for more details",
    )
    run_parser.add_argument(
        '--directory',
        '-d ',
        help='Working directory, must include the peripherals.json and scenario.json files. Default is current working directory',
        default=getcwd()
    )

    run_parser.add_argument(
        '--sudo',
        '-s ',
        help='Run in sudo mode => FW can write to read-only registers. This should usually be used for testing low-level drivers',
        action='store_true',
        default=False
    )

    run_parser.add_argument(
        '--gdb',
        '-g ',
        help='Opens a GDB port for debugging the FW on port 5555. The FW will not start running until the GDB client connects.',
        action='store_true',
        default=False
    )

    run_parser.add_argument(
        '--version',
        '-v ',
        help='Jumper sdk version.',
        action='store_true',
        default=False
    )

    run_parser.add_argument(
        '--trace',
        '-t ',
        action='store_true',
        help='Prints a trace report to stdout, this can be used with --trace-dest to forward it to a file.',
        default=False
    )

    run_parser.add_argument(
        '--trace-dest',
        type=str,
        help=
        """
        Forwards the trace report to a destination file. This MUST be used -t with this flag to make it work.
        To print to stdout, just hit -t.
        """,
        default="",
    )

    run_parser.add_argument(
        '--uart',
        '-u ',
        action='store_true',
        default=False,
        help='Forwards UART prints to stdout, this can be used with --uart-dest to forward it to a file.'
    )

    run_parser.add_argument(
        '--uart-dest',
        type=str,
        help=
        """
        Forwards UART prints to a destination file. This MUST be used -u with this flag to make it work.
        To print to stdout, just hit -u.
        """,
        default="",
    )

    run_parser.add_argument(
        "--gpio",
        help=
        """
        Prints GPIO events to stdout.
        """,
        action='store_true',
        default=False,
    )

    run_parser = subparsers.add_parser(
        'ble',
        help='Creates a virtual HCI device (BLE dongle) for regular Linux/Bluez programs to communicate with virtual devices'
    )

    args = parser.parse_args()

    if args.command == 'run':
        if args.version:
            print ("v" + jumper_current_version)
            exit(0)
        if not args.bin:
            print("jumper run: error: argument --bin/-b is required")
            exit(1)
        run(args)

    if args.command == 'ble':
        hci()


if __name__ == '__main__':
    main()



"""
:copyright: (c) 2017 by Jumper Labs Ltd.
:license: Apache 2.0, see LICENSE.txt for more details.
"""

from time import sleep
from jemu_mem_peripheral import JemuMemPeripheral
import threading


class JemuSudo(JemuMemPeripheral):
    _id = None
    _jemu_connection = None
    _peripheral_type = None
    _state = None

    _STOP_AFTER_COMMAND = "stop_after"
    _START_COMMAND = "resume_running"
    _COMMAND = "command"
    _NANO_SECONDS = "nanoseconds"
    _TYPE_STRING = "type"
    _PERIPHERAL_ID = "peripheral_id"
    _PERIPHERAL_TYPE = "peripheral_type"
    _STOPPED = "stopped"
    _RESUMED = "resumed"
    _COMMAND_SET_TIMER = "set_timer"
    _MESSAGE_ID = "message_id"
    _GET_STATE_COMMAND = "get_state"
    _CANCEL_STOP_ON_TICK = "cancel_stop"
    _GET_DEVICE_TIME = "get_device_time"

    def __init__(self, jemu_connection, id, peripheral_type):
        JemuMemPeripheral.__init__(self, jemu_connection, id, peripheral_type, None)
        self._id = id
        self._timer_id_counter = 0
        self._peripheral_type = peripheral_type
        self._jemu_connection = jemu_connection
        self._jemu_connection.register(self.receive_packet)
        self._timer_id_callback_dict = {}
        self._state_lock = threading.RLock()
        self._state = None
        self._device_time_lock = threading.RLock()
        self._device_time = None
        self._resumed_packet_received = threading.Event()
        self._stopped_packet_received = threading.Event()

    def _generate_json_command(self, command):
        return {
            self._TYPE_STRING: self._COMMAND,
            self._PERIPHERAL_ID: self._id,
            self._COMMAND: command,
            self._PERIPHERAL_TYPE: self._peripheral_type
        }

    def stop_after_ns(self, nanoseconds):
        json_command = self._generate_json_command(self._STOP_AFTER_COMMAND)
        json_command[self._NANO_SECONDS] = nanoseconds
        self._jemu_connection.send_json(json_command)

    def run_for_ns(self, nanoseconds):
        if self.get_state() == 'running':
            self._stopped_packet_received.clear()
            self.stop_after_ns(0)
            while not self._stopped_packet_received.is_set():
                self._stopped_packet_received.wait(0.2)

        if nanoseconds > 0:
            self._stopped_packet_received.clear()
            self.stop_after_ns(nanoseconds)
            self.resume()
            while not self._stopped_packet_received.is_set():
                self._stopped_packet_received.wait(0.2)

    def resume(self):
        if self.get_state() == 'paused':
            self._resumed_packet_received.clear()
            self._jemu_connection.send_json(self._generate_json_command(self._START_COMMAND))
            self._resumed_packet_received.wait()

    def wait_until_stopped(self):
        self._stopped_packet_received.clear()
        while not self._stopped_packet_received.is_set():
            self._stopped_packet_received.wait(0.2)

    def cancel_stop(self):
        self._jemu_connection.send_json(self._generate_json_command(self._CANCEL_STOP_ON_TICK))

    def receive_packet(self, jemu_packet):
        if jemu_packet[self._TYPE_STRING] == self._STOPPED:
            self._stopped_packet_received.set()

        elif jemu_packet[self._TYPE_STRING] == self._RESUMED:
            self._resumed_packet_received.set()

        elif 'type' in jemu_packet and jemu_packet["type"] == "timer_id":
            cur_id = jemu_packet["value"]
            for id_from_dict in self._timer_id_callback_dict:
                if id_from_dict == int(cur_id):
                    self._timer_id_callback_dict[int(cur_id)]()

        elif 'type' in jemu_packet and jemu_packet["type"] == "emulator_state":
            with self._state_lock:
                self._state = jemu_packet["value"]

        elif 'type' in jemu_packet and jemu_packet["type"] == "device_time":
            with self._device_time_lock:
                self._device_time = jemu_packet["value"]

    def set_timer(self, ticks, callback):
        self._timer_id_callback_dict.update({self._timer_id_counter: callback})
        json_command = self._generate_json_command(self._COMMAND_SET_TIMER)
        json_command[self._NANO_SECONDS] = ticks
        json_command[self._MESSAGE_ID] = self._timer_id_counter
        self._timer_id_counter += 1
        self._jemu_connection.send_json(json_command)

    def get_state(self):
        self._state = None
        json_command = self._generate_json_command(self._GET_STATE_COMMAND)
        self._jemu_connection.send_json(json_command)
        state = self._state
        while state is None:
            sleep(0.01)
            with self._state_lock:
                state = self._state
        return self._state

    def _get_device_time(self):
        stop = False
        while not stop:
            sleep(0.01)
            with self._device_time_lock:
                if self._device_time is not None:
                    stop = True
        return self._device_time

    def get_device_time_ns(self):
        self._device_time = None
        json_command = self._generate_json_command(self._GET_DEVICE_TIME)
        self._jemu_connection.send_json(json_command)
        return self._get_device_time()


